/*
 *****  loadsig_mcs.c written by Mark C Spaeth *****
 * Loads in Hspice and Finesim simulation files using 9601 format. 
 * Improves the original loadsig.c function by Michael H Perrott 
 * with cleaner code, faster load times, and better support of sweeps.
 * This software comes with no warranty or support, and is distributed
 * under the GNU General Public License as described below.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the MxFree
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <float.h>
#include "mat.h"
#include "mex.h"


#define VOLT 1 
#define FREQ 2
#define CURR 8

#define MAX_FNAME 1024
#define MAX_SNAME 128

struct blkll {
  int len, eob;
  struct blkll *next;
};

/*
Independent:
 1 = Transient (TIME)
 2 = AC Sweep (HERTZ)
 3 = DC Sweep (VOLTS)

TYPES:
 1 = Voltage (Complex in AC sim)?
 2 = Voltage (Magnitude)
 5 = Voltage (Phase)
 8 = Current
78 = Parameter
79 = LX Parameter 
80 = LV Parameter 
*/

int head[4];

size_t (*fread_dummy)(void *data, size_t size, size_t n, FILE *fp);
size_t fread_err(void *data, size_t size, size_t n, FILE *fp);
size_t fread_we(void *data, size_t size, size_t n, FILE *fp);
void addblk(struct blkll **blks, int len);
int mblklen(struct blkll **blks);
void fixname(char *signame);
int validateblock(FILE *fp);
int readheadblock(char *data, FILE *fp);
int readdatablock(float *data, FILE *fp);
int line;

void mexFunction(int nret, mxArray **ret, int argc, const mxArray **argv) {
  FILE *fp;
  char fname[MAX_FNAME];
  char *header,*hptr,*signame;
  int len,maxrows,swptyp;
  int i,j,k,nind,ndep,nswp,ntot,ncols,swplen;
  /* int *sigcols; */
  struct blkll *blks, *curblk, *tmpblk;
  const char *fields[8] = {"name","data"};
  double **matdata;
  float *blkdata,*blkptr,*blkend;
  int *matptr;
  mxArray *mxtmp;

  if (argc!=1) 
    mexErrMsgTxt("Error: Missing filename\nUsage: x=loadsig('filename')");
  if (!mxIsChar(argv[0]))
    mexErrMsgTxt("Error: Filename must be a string");
  if (nret!=1)
    mexErrMsgTxt("Error: Missing output variable");

  mxGetString(argv[0],fname,MAX_FNAME-1);

  if ((fp=fopen(fname,"rb"))==NULL)
    mexErrMsgTxt("Error: Unable to open specified file");

  /* Test for wrong endianness */
  fread_dummy=fread;
  fread_err(head,4,4,fp);
  if (head[0]==0x4000000 && head[2]==0x40000000) {
    mexPrintf("Warning: Possible wrong endianness in data\n");
    fread_dummy=fread_we;
  }

  blks=NULL;
  rewind(fp);

  /* Validate header blocks */
  while (1) {
    if (len=validateblock(fp))
      addblk(&blks,len);
    if (len<=0) break;
  }
  if (len==0) mexErrMsgTxt("Error: Hit EOF before end of header");

  /* Validate data blocks */
  while (len=validateblock(fp))
    addblk(&blks,len);

  /*
  tmpblk=blks;
  i=0;
  while (len=mblklen(&tmpblk))
    fprintf(stderr,"%d %04x %08x %08x\n",i++,len,blks,tmpblk);

    MBLKLEN should return ncols * nrows + nswp + 1 (end of block trailer), check this after counting variables.
  */

  curblk=blks;
  len=mblklen(&curblk);
  /*  fprintf(stderr,"hlen: %04x\n",len); */

  /* Read ascii header into a structure */
  /* fprintf(stderr,"Alloc ascii header\n"); */
  rewind(fp);
  if ((header = (char *)mxCalloc(len+1,sizeof(char))) == NULL) 
    mexErrMsgTxt("Error: mxCalloc failed (file header)");
  header[len]=0;
  hptr=header;

  /* fprintf(stderr,"Reading ascii header\n"); */
  curblk=blks;
  while (1) {
    /* fprintf(stderr,"hptr: %08x %d\n",hptr,curblk->eob); */
    hptr+=readheadblock(hptr,fp);
    if (curblk->eob) {
      curblk=curblk->next;
      break;
    }
    curblk=curblk->next;
  }
  /* fprintf(stderr,"%08x %08x %08x\n",header,header+len,hptr); */
  
  /* Remove false EOS's if necessary */
  /* fprintf(stderr,"Removing EOFs\n"); */
  for (i=0;i<len;i++) 
    if (header[i]==0) header[i]=' ';

  if (strncmp(&header[0x10],"9601    ",8))
    mexErrMsgTxt("Error: Output file not in 9601 post format");

  if (sscanf(header,"%4d%4d%4d",&nind,&ndep,&nswp)!=3)
    mexErrMsgTxt("Error: Can't parse signal enumeration");
  if (nswp>1) 
    mexErrMsgTxt("Error: File contains >1 sweep?!");

  ntot=nind+ndep+nswp;
  /*  mexPrintf("nind = %d, ndep = %d, nswp = %d\n",nind,ndep,nswp); */

  /* Find out if AC to adjust for complex auto probes */
  hptr=strtok(header+0x108," \t\n");
  if ((swptyp=atoi(hptr))==0)
    mexErrMsgTxt("Error: Can't read sweep type");
  ncols=nind+ndep+((swptyp==2)?(nind-1):0); 

  /* Initialize array w/ # cols per signal -- to handle complex data */
  /* Don't do this for now, since in ac sweep swptyp=2, all nind but first are complex? */
  /*
  sigcols=(int *)mcCalloc(ntot,sizeof(int));
  for (i=0;i<ntot;i++)
    sigcols(i)=1;
  */
   
  /* Parse sweep lengths */
  /* fprintf(stderr,"Parsing sweep lengths\n"); */
  if (nswp) {
    if (sscanf(&header[0xb8],"%4d",&swplen)!=1) 
      mexErrMsgTxt("Error: Can't parse sweep length");
  } else swplen=1;


  /* Get max data block size to determine # rows */
  /*  fprintf(stderr,"Getting # rows\n"); */
  tmpblk=curblk;
  maxrows=0;
  while (len=mblklen(&tmpblk)) {
    len>>=2;
    len-=(nswp+1);
    /* fprintf(stderr,"%d %d %d %d\n",len,nind+ndep,ncols,len%(ncols)); */
    if (len%(ncols))
      mexErrMsgTxt("Invalid data block size");
    len/=(ncols);
    /* mexPrintf("# Rows = %d\n",len); */
    maxrows=(len>maxrows)?len:maxrows;
  }

  /* Start of signal names -- may be a more robust way of finding this */
  hptr=header+0x108+0x08*(ndep+nind);

  /* Copy field names into matlab structure */
  /* fprintf(stderr,"Copying signal names\n"); */
  matdata=(double **)mxCalloc(ntot,sizeof(double *));
  matptr=(int *)mxCalloc(ntot,sizeof(int));
  ret[0]=mxCreateStructMatrix(ntot,1,2,fields);
  for (i=0;i<ntot;i++) {
    if ((signame=strtok(hptr," \t\n"))==NULL)
      mexErrMsgTxt("Missing signal name in header");
    else {
      fixname(signame);
      /* fprintf(stderr,"%s\n",signame); */
      mxtmp=mxCreateString(signame);
      mxSetField(ret[0],i,fields[0],mxtmp);
      if ((swptyp==2) && i && (i<nind)) 
	mxtmp=mxCreateDoubleMatrix(maxrows,swplen,mxCOMPLEX);
      else if (nswp && (i==ntot-1))
	mxtmp=mxCreateDoubleMatrix(1,swplen,mxREAL);
      else
	mxtmp=mxCreateDoubleMatrix(maxrows,swplen,mxREAL);
      mxSetField(ret[0],i,fields[1],mxtmp);
      matdata[i]=(double *)mxGetPr(mxtmp);
    }
    hptr=NULL;
  }

  /* Done with header, free the memory */
  mxFree(header);

  /* Read and parse data blocks */
  for (i=0;i<swplen;i++) {
    tmpblk=curblk;
    len=mblklen(&tmpblk);
    if ((blkdata=(float *)mxCalloc(len,sizeof(float)))==NULL)
      mexErrMsgTxt("Error: mxCalloc failed (data block)");
    blkptr=blkdata;

    /* Read 'complete' multi-block data block */
    while (1) {
      blkptr+=readdatablock(blkptr,fp);
      if (curblk->eob) {
	curblk=curblk->next;
	break;
      }
      curblk=curblk->next;
    }
    
    /* Set pointers to start of rows */
    for (j=0;j<nind+ndep;j++)
      if ((swptyp==2) && j && (j<nind)) 
	matptr[j]=2*i*maxrows;
      else
	matptr[j]=i*maxrows;

    line=0;
    /* Grab sweep index, if there's sweeps */
    blkend=blkptr-1;
    blkptr=blkdata;
    if (nswp) {
      matdata[ntot-1][i]=(double)*blkptr;
      blkptr++;
    }
    
    /* Sort data block into matlab arrays */
    while(blkptr<blkend) {
      /* fprintf(stderr,"%08x %08x\n",blkptr,blkend); */
      /* fprintf(stderr,"%4d : %4d %e %08x\n",line++,(blkptr-blkdata)/4,*blkptr,blkptr);*/
      for (j=0;j<nind+ndep;j++) {
	k=((swptyp==2) && j && (j<nind)) ? 2 : 1;
	while(k--) {
	  matdata[j][matptr[j]]=*blkptr;
	  blkptr++;
	  matptr[j]++;
	}
      }
    }

    /* All done with this data block */
    mxFree(blkdata);
  }
    

  /* Close file */
  fclose(fp);

  /* fprintf(stderr,"Freeing structs\n"); */
  curblk=blks;
  while (curblk) {
    tmpblk=curblk;
    curblk=tmpblk->next;
    /* fprintf(stderr,"%08x %08x\n",tmpblk,curblk); */
    mxFree(tmpblk);
  }
  mxFree(matdata);
  mxFree(matptr);
}


/* Fread with EOF checking to make the rest of the code cleaner */
size_t fread_err(void *data, size_t size, size_t n, FILE *fp) {
  if (fread_dummy(data,size,n,fp)!=n) {
    mexPrintf("Error: Unexpected EOF at 0x%08x\n",ftell(fp));
    mexErrMsgTxt("Unexpected EOF");
  }
  return n;
}


/* Wrong endian read (assume a 4-byte size) */
size_t fread_we(void *data, size_t size, size_t n, FILE *fp) {
  unsigned char *cptr,c;
  size_t i;

  if (fread(data,size,n,fp)!=n) return 0;

  for (i=0;i<=n;i++) {
    cptr=(unsigned char *)data+(4*i);
    c=cptr[0];
    cptr[0]=cptr[3];
    cptr[3]=c;
    c=cptr[1];
    cptr[1]=cptr[2];
    cptr[2]=c;
  }
  return n;
}


/* Check the validity of a block of data 
   Header: 0x04, <# entries>, 0x04 (data size?), <length> 
   Trailer: <length>

   Last 2 entries in full header = 0x23252624, 0x20202020
   Last entry in sweep block     = 0x7149f2ca  ( ie. 1e30 as a 32-bit float ) 

   Return = Length of block, negative if last block, 0 on an error */
int validateblock(FILE *fp) {

  /* Return if EOF reading header -- could be real EOF, so don't use fread_err */
  if (fread_dummy(head,4,4,fp)!=4)
    return 0;

  if ((head[0]!=0x04)||(head[2]!=0x04)) {
    mexPrintf("Error: Invalid block header at 0x%08x\n",ftell(fp));
    mexErrMsgTxt("Bad header");
  }

  /*  mexPrintf("Length: %08x\n",head[3]); */

  /* Skip until 2 entries before the end of the block */
  fseek(fp,head[3]-8,SEEK_CUR);
  fread_err(head, 4, 3, fp);

  /* Check trailer vs block length */
  if (head[2]!=head[3]) {
    mexPrintf("Error: Invalid block trailer at 0x%08x\n",ftell(fp));
    mexErrMsgTxt("Bad trailer");
  }

  /* Check if end of header block */
  if ((head[0]==0x23252624)&&(head[1]==0x20202020))
    return -head[3];

  /* Check if end of sweep block */
  if (head[1]==0x7149f2ca)
    return -head[3];

  return head[3];
}  


/* Read in a header block, skipping the header / trailer
   minimal error checking, as the block should be verified by now */
int readheadblock(char *data, FILE *fp) {
  int len;

  fread_dummy(head,4,4,fp);
  len=fread(data,1,head[3],fp);
  fread_dummy(head,1,4,fp);
  
  /* fprintf(stderr,"%d %d %d\n",head[0],head[3],len); */

  if ((head[0]!=head[3])||(head[0]!=len)) {
    mexPrintf("Error reading header block at 0x%08x\n",ftell(fp));
    mexErrMsgTxt("Bad header block");
  }
  
  return len;
}


/* Read in a data block, skipping the header / trailer
   minimal error checking, as the block should be verified by now */
int readdatablock(float *data, FILE *fp) {
  int len;

  /* fprintf(stderr,"%08x ",ftell(fp)); */
  fread_dummy(head,4,4,fp);
  /* fprintf(stderr,": %08x : %04x : ",ftell(fp),head[1]); */
  len=fread_dummy(data,4,head[3]/4,fp);
  /* fprintf(stderr,": %08x\n",ftell(fp)); */
  fread_dummy(head,1,4,fp);
  /* fprintf(stderr,"%d %d %d\n",head[0],head[3],len); */

  if ((head[0]!=head[3])||(head[3]!=len*4)) {
    mexPrintf("Error reading data block at 0x%08x\n",ftell(fp));
    mexErrMsgTxt("Bad data block");
  }
  
  return len;
}

  
/* Adds a block to the and of the linked list structure */
void addblk(struct blkll **blks, int len) {
  struct blkll *tmp, *curr;
  int i=0;

  if ((tmp = (struct blkll *)mxCalloc(1,sizeof(struct blkll)))==NULL)
    mexErrMsgTxt("mxCalloc failed!");
  tmp->len=abs(len);
  tmp->eob=(len<0);
  tmp->next=NULL;

  curr=(*blks);
  if (!curr)
    (*blks)=tmp;
  else {
    while (curr->next)
      curr=curr->next;
    curr->next=tmp;
  }
}


/* Returns the length of the current multi-block block, and advances the pointer to the next one */
int mblklen(struct blkll **blks) {
  int len=0;

  if ((*blks)==NULL) {
    return 0;
  } else {
    while (!((*blks)->eob)) {
      len+=(*blks)->len;
      (*blks)=(*blks)->next;
    }
    len+=(*blks)->len;
    (*blks)=(*blks)->next;
  }
  return len;
}
      

/* Replaces 'bad' characters in signal names with '_'
   and truncates trailing '_' for hspice/finseim compatibility */
 void fixname(char *signame) {
   int i;
   char c;

   for (i=0;i<strlen(signame);i++) {
     c=signame[i];
     if ( (c<'0') || (c>'z')  || (c=='=') ||
	 ((c>'9') && (c<'<')) || 
	 ((c>'>') && (c<'A')) ||
         ((c>'Z') && (c<'a')))
       if (i<strlen(signame)-1)
	 signame[i]='_';
       else
	 signame[i]='\0';
   }
 }

       
